//
//  PCKey.h
//  PayControl
//
//  Created by Roman Kolesnichenko on 12.08.16.
//  Copyright © 2016 SafeTech. All rights reserved.
//

#import "PCError.h"

@class PCTransaction;
@class PCQRCode;

static const int MIN_D = 6;
static const int MAX_D = 10;
static const int HASH_LEN = 32 * 2;

/**  */
typedef NS_ENUM(NSUInteger, PCCodeType)
{
    /** */
    PCCodeTypeHMAC = 1,
    /** */
    PCCodeTypeAUTH = 2
};

/** Тип ключа */
typedef NS_ENUM(NSUInteger, PCKeyType)
{
    /** Тип ключа Regular */
    PCKeyTypeRegular = 1,
    /** Тип ключа DSS (КриптоПРО) */
    PCKeyTypeDSS = 2
};

/** Предпочитаемый тип подтверждения транзакции */
typedef NS_ENUM(NSUInteger, PCConfirmationType)
{
    /** Транзакция подтверждается в офлайне */
    PCConfirmationTypeOffline = 1,
    /** Транзакция подтверждается в онлайне */
    PCConfirmationTypeOnlineURL = 2,
    /** Транзакция подтверждается через Push уведомление */
    PCConfirmationTypeOnlinePush = 3
};

/** Объект ключа. */
@interface PCKey : NSObject

/// ****************************************************************************
/// @name Методы класса
/// ****************************************************************************

/** Создание экзепляра ключа из объекта QR-кода (PCKeyQRCode). 
 @param qrCode объект класса PCQRCode.
 @return Возвращает экзепляр ключа либо nil в случае ошибки.
 */
+ (nullable instancetype)keyFromQRCode:(PCQRCode*)qrCode;

/** Создание экзепляра ключа из строки с вадидными данными QR-кода ключа. 
 @param qrCodeString строка с вадидными данными QR-кода ключа.
 @return Возвращает экзепляр ключа либо nil в случае ошибки.
 */
+ (nullable instancetype)keyFromQRCodeString:(NSString*)qrCodeString;

/** Создание экзепляра ключа из строки с вадидными данными в формате XML. 
 @param xmlString строка с вадидными данными ключа в формате XML.
 @return Возвращает экзепляр ключа либо nil в случае ошибки.
 */
+ (nullable instancetype)keyFromXMLString:(NSString*)xmlString;

/** Получение экзепляра ключа по его идентификатору в хранилище.
 @param keyID строка с индентификатором ключа.
 @return Возвращает экзепляр ключа либо nil в случае ошибки.
 */
+ (nullable instancetype)keyFromStorage:(NSString *)keyID;

#pragma mark -

/// ****************************************************************************
/// @name Методы
/// ****************************************************************************

/**
 Объединение со второй частью ключа
 @param keyPart - Вторая часть ключа
 @returns error - ошибка. Если error = nil, то добавление второй части ключа завершилось успешно
*/
- (nullable PCError *)combineKey:(NSString *)keyPart;

/**
 Сохранение ключевой информации (полный ключ)
 @param keyName - Имя ключевой информации
 @param  password - Пароль для ключевой информации
 @return error - ошибка. Если error = nil, то сохранение завершилось успешно
 */
- (nullable PCError*)storeWithName:(NSString*)keyName password:(NSString *)password;

/**
 Сохранение ключевой информации в хранилище без указания пароля
 Функция необходима для сохранения ранее импортированных ключей
 без предъявления текущего пароля
 Для последующего доступа к ключам необходим старый пароль
 Валидный handle не требуется
 
 @param keyName - Имя ключевой информации
 @return error - ошибка. Если error = nil, то сохранение завершилось успешно
 */
- (nullable PCError*)storeWithName:(NSString*)keyName;

/**
 Предъявление пароля для получения Handle ключевой информации. В случае успешного ввода пароля Handle становится валидным для использования при генерации кода подтверждения
 @param password - пароль
 @return error - ошибка. Если error = nil, то предъявление пароля прошло успешно.
 */
- (nullable PCError *)submitPassword:(NSString *)password;

/**
 Изменение пароля для ключевой информации
 @param oldPassword - Текущий пароль для ключевой информации
 @param newPassword - Новый пароль для ключевой информации
 @return error - ошибка. Если error = nil, то обновление пароля прошло успешно.
 */
- (nullable PCError *)updatePassword:(NSString *)oldPassword newPassword:(NSString *)newPassword;

/**
 Изменение имени для ключевой информации
 @param keyName - Имя ключевой информации
 @return error - ошибка. Если error = nil, то обновление имени прошло успешно.
 */
- (nullable PCError *)updateKeyName:(NSString *)keyName;

/**
 Получение интервала дискретизации времени
 @param codeType тип, для которого нужно получить интервал дискретизации времени. см. PCCodeType
 @return значение интервала дискретизации времени. По умолчанию равно 180 секундам.
 */
- (NSTimeInterval)getCodeExpirationInterval:(PCCodeType)codeType;

/**
 Установка интервала дискретизации времени
 @param interval значение интервала дискретизации времени
 @param codeType тип, для которого изменяется интервал дискретизации времени см. PCCodeType
 */
- (void) setCodeExpirationInterval:(NSTimeInterval)interval codeType:(PCCodeType)codeType;


#pragma mark -

/** Освободить хэндлы ключей текущего объекта */
- (void)freeHandle;

#pragma mark - ECDSA

/** Генерация ключевой пары. Создание двух объектов – открытого и закрытого ключа. */
- (nullable PCError *)generateKeyPair;

/** Очистка ключевой пары. Удаление (очистка) у объекта открытого и закрытого ключей. */
- (nullable PCError *)clearKeyPair;

#pragma mark -

/* Формирование авторизационного кода для получения данных транзакции */
- (nullable NSString*)calculateAuthCodeForGetTransactionData:(NSString*)transactionID timeShift:(NSTimeInterval)T;
/* Формирование авторизационного кода для отправки отказа транзакции */
- (nullable NSString*)calculateAuthCodeForDeclineTransaction:(NSString*)transactionID declineReason:(NSString*)declineReason timeShift:(NSTimeInterval)T;
/* Формирование авторизационного кода для отпрвки отпечатка утсройства */
- (nullable NSString*)calculateAuthCodeForFingerPrint:(NSString*)fingerprint timeShift:(NSTimeInterval)T;
/* Формирование авторизационного кода для получения списка транзаций */
- (nullable NSString*)calculateAuthCodeForGetTransactions:(NSString*)status timeShift:(NSTimeInterval)T;
/* Формирование авторизационного кода для регистрации устройства */
- (nullable NSString*)calculateAuthCodeForRegistrationDevice:(NSString*)deviceID timeShift:(NSTimeInterval)T;
/* Формирование кода для регистрации публичного ключа */
- (nullable NSString*)calculateAuthCodeForRegistrationPublicKey:(NSTimeInterval)timeShift;
/* Формирование кода для подтверждения транзации */
- (nullable NSString*)generateConfirmationCode:(PCTransaction*)transaction fingerprint:(NSString*)fingerprint timeShift:(NSTimeInterval)T D:(NSUInteger)D;
/* Формирование кода для отклонения транзации */
- (nullable NSString*)generateDeclineCode:(PCTransaction*)transaction fingerprint:(NSString*)fingerprint timeShift:(NSTimeInterval)T D:(NSUInteger)D;
/* Формирование подписи */
- (nullable NSString*)calculateSignature:(PCTransaction*)transaction fingerprint:(NSString*)fingerprint timeShift:(NSTimeInterval)T D:(NSUInteger)D;


#pragma mark -
/// ****************************************************************************
/// @name Свойства
/// ****************************************************************************

/**
 Проверка на возможность использования Handle
 */
@property(nonatomic, readonly) BOOL hasValidHandle;

// Ключевая информация теперь представляет из себя 2 ключа (один для транзакций (0, length / 2),
// второй для кода подтверждения (length / 2, length). Теперь отдельно нужно иметь указатель на handle первого и второго ключей (hmac и auth)

/** Handle ключа для выработки кода подтверждения транзакции */
@property(nonatomic, readonly, nullable) NSString* handleHMAC;

/** Handle ключа для выработки кода аутентификации */
@property(nonatomic, readonly, nullable) NSString* handleAuth;

/** Handle приватного ключа ECDSA */
@property(nonatomic, readonly) NSString* handlePrivate;

/** Идентификатор пользователя PayControl */
@property(nonatomic, strong, nullable) NSString *userID;

/** Наименование ключа */
@property(nonatomic, strong, nullable) NSString *keyName;

/** Дата истечения срока действия ключа */
@property(nonatomic, strong, nullable) NSDate *expirationDate;

/** Дата истечения срока действия ключа в виде строки "dd.MM.yyyy" */
@property(nonatomic, strong, nullable) NSString *expirationDateString;

/** Дата истечения срока действия ключа в виде строки в локализованном формате даты */
@property(nonatomic, readonly) NSString *localizedExpirationDateString;

/** Признак того, что у ключа истек срок действия */
@property(nonatomic, readonly) BOOL isExpiredKey;

/** Идентификатор ключа в хранилище PayControl */
@property(nonatomic, strong, nullable) NSString *keyID;

/** Тип ключа: DSS или Regular см. PCKeyType */
@property(nonatomic, assign) PCKeyType keyType;

/** Идентификатор прикладной системы */
@property(nonatomic, strong, nullable) NSString *systemID;

/** Флаг необходимости использования отпечатка устройства */
@property(nonatomic, assign) BOOL isWithFingerprint;

/** Строка с URL для подтверждения Online */
@property(nonatomic, strong, nullable) NSString *onlineConfirmURL;

/** Строка с URL PC Interaction Server */
@property(nonatomic, strong, nullable) NSString *interactionExternalURL;

// вид алгоритма шифрования, возможно понадобиться в будущем, сейчас туда записывается и сохраняется в базе константа - AES_ECB_SHA256 */
//@property (nonatomic, strong) NSString *encryptingAlgorithm;

/**
 Проверка на полноту ключа.
 @return Код ошибки. Если код ошибки == PC_ERROR_OK, то ключ заполнен полностью. Если код ошибки == PC_ERROR_KEY_IS_NOT_FULL, то необходимо объединение со второй частью ключа
 */
@property(nonatomic, readonly, nullable) PCError *isCompleted;


/** Токен устройства, зарегистрированного на PC IS при помощи метода doRegisterDevice */
@property(nonatomic, strong, nullable) NSString *registeredPushToken;

/** Возвращает признак, имеет ли ключ сформированную ключевую пару */
@property(nonatomic, readonly) BOOL hasKeyPair;

/** возвращает признак, имеет ли ключ валидный хэндл для формирования подписи. */
@property(nonatomic, readonly) BOOL hasValidSignHandle;

/** возвращает публичную часть ключа. */
@property(nonatomic, nullable) NSString *publicKey;

@end
